/*
20080912
D. J. Bernstein
Public domain.
*/

#define X_16 \
  X(000); X(001); X(002); X(003); X(004); X(005); X(006); X(007);\
  X(010); X(011); X(012); X(013); X(014); X(015); X(016); X(017)

static void add(unsigned int h[17],const unsigned int c[17])
{
  unsigned int j;
  unsigned int u;
  u = 0;
  for (j = 0;j < 17;++j) { u += h[j] + c[j]; h[j] = u & 255; u >>= 8; }
}

static void squeeze(unsigned int h[17])
{
  unsigned int j;
  unsigned int u;
  u = 0;
  for (j = 0;j < 16;++j) { u += h[j]; h[j] = u & 255; u >>= 8; }
  u += h[16]; h[16] = u & 3;
  u = 5 * (u >> 2);
  for (j = 0;j < 16;++j) { u += h[j]; h[j] = u & 255; u >>= 8; }
  u += h[16]; h[16] = u;
}

static const unsigned int minusp[17] = {
  5, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 252
} ;

static void freeze(unsigned int h[17])
{
  unsigned int horig[17];
  unsigned int j;
  unsigned int negative;
  //for (j = 0;j < 17;++j) horig[j] = h[j];
#define X(j) horig[j] = h[j]
  X_16;
  X(16);
#undef X
  add(h,minusp);
  negative = -(h[16] >> 7);
  for (j = 0;j < 17;++j) h[j] ^= negative & (horig[j] ^ h[j]);
}

static void mulmod(unsigned int h[17],const unsigned int r[17])
{
  unsigned int hr[17];
  unsigned int i;
  unsigned int j;
  unsigned int u;

  //for (i = 0;i < 17;++i)
  {
#define X(i) \
    u = 0;\
    for (j = 0;j <= i;++j) u += h[j] * r[i - j];\
    for (j = i + 1;j < 17;++j) u += 320 * h[j] * r[i + 17 - j];\
    hr[i] = u
    X_16;
    X(16);
#undef X
  }
  for (i = 0;i < 17;++i) h[i] = hr[i];
  squeeze(h);
}

int crypto_onetimeauth(unsigned char *out,const unsigned char *in,unsigned long long inlen,const unsigned char *k)
{
  unsigned int j;
  unsigned int r[17];
  unsigned int h[17];
  unsigned int c[17];

  r[0] = k[0];
  r[1] = k[1];
  r[2] = k[2];
  r[3] = k[3] & 15;
  r[4] = k[4] & 252;
  r[5] = k[5];
  r[6] = k[6];
  r[7] = k[7] & 15;
  r[8] = k[8] & 252;
  r[9] = k[9];
  r[10] = k[10];
  r[11] = k[11] & 15;
  r[12] = k[12] & 252;
  r[13] = k[13];
  r[14] = k[14];
  r[15] = k[15] & 15;
  r[16] = 0;

  //for (j = 0;j < 17;++j) h[j] = 0;
#define X(j) h[j] = 0
  X_16;
  X(16);
#undef X

  unsigned long long i = 0;
  while( inlen - i > 16 ) {
#define X(j) c[j] = in[i + j]
    X_16;
#undef X
    c[16] = 1;
    i += 16;
    add(h,c);
    mulmod(h,r);
  }
#define X(j) \
  c[j] = (i + j < inlen) ? in[i + j] \
  : (i + j == inlen) ? 1 : 0
  X_16;
  X(16);
#undef X
  add(h,c);
  mulmod(h,r);

  freeze(h);

  //for (j = 0;j < 16;++j) c[j] = k[j + 16];
#define X(j) c[j] = k[j + 16]
  X_16;
#undef X
  c[16] = 0;
  add(h,c);
  for (j = 0;j < 16;++j) out[j] = h[j];
  return 0;
}

#define MAXTEST_BYTES 10000
#define CHECKSUM_BYTES 4096
#define TUNE_BYTES 15 /*36*/

#define crypto_onetimeauth_BYTES 16
#define crypto_onetimeauth_KEYBYTES 32

#define Z10 0,0,0,0,0,0,0,0,0,0
#define Z100 Z10,Z10,Z10,Z10,Z10,Z10,Z10,Z10,Z10,Z10 
#define Z1000 Z100,Z100,Z100,Z100,Z100,Z100,Z100,Z100,Z100,Z100 
#define Z10000 Z1000,Z1000,Z1000,Z1000,Z1000,Z1000,Z1000,Z1000,Z1000,Z1000 

static unsigned char h[crypto_onetimeauth_BYTES] =
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };
static unsigned char m[MAXTEST_BYTES] = { Z10000 };
static unsigned char k[crypto_onetimeauth_KEYBYTES] =
{ 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

void doit(void)
{
  crypto_onetimeauth(h,m,TUNE_BYTES,k);
  // crypto_onetimeauth_verify(h,m,TUNE_BYTES,k);
}

int
main(void)
{
  doit();
  return 0;
}
