// Test linearization of parabola

#include"verasco.h"

void parabola(int lb, int ub, int c) { 
  int x=any_int(), r;

  // check test validity !
  verasco_assert(lb <= ub);

  // initialization of inputs
  verasco_assume(lb <= x && x <= ub);

  // TESTS
  r=c*x*x;

  // Result check: best interval approx
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub); 
    } else if (ub <= 0) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else {
      verasco_assert(0 <= r);
      if (-lb <= ub) {
        verasco_assert(r <= c*ub*ub);
      } else {
        verasco_assert(r <= c*lb*lb);
      }
    }
  } else {
    if (0 <= lb) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else if (ub <= 0) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub);    
    } else {
      verasco_assert(r <= 0);
      if (-lb <= ub) {
        verasco_assert(c*ub*ub <= r);
      } else {
        verasco_assert(c*lb*lb <= r);      
      }
    }
  }

  // Result check: best poly approx on the "intern" side 
  if (c >= 0) {
    verasco_assert(r <= c*((ub+lb)*x - lb*ub));
  } else { 
    verasco_assert(c*((ub+lb)*x - lb*ub) <= r);
  }

  // Result check: on the "extern" side, tangent that 
  // generalizes the interval bound
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*(2*x-lb) <= r);
    } else if (ub <= 0) {
      verasco_assert(c*ub*(2*x-ub) <= r);
    } // else 0 <= r
  } else {
    if (0 <= lb) {
      verasco_assert(r <= c*lb*(2*x-lb));
    } else if (ub <= 0) {
      verasco_assert(r <= c*ub*(2*x-ub));
    } // else 0 <= r
  }

}

void test_parabola() { 
  // 0 <= c 
  parabola(0,10,2);
  parabola(10,101,2);
  parabola(-101,0,3);
  parabola(-101,-10,4);
  parabola(0,10,5);
  parabola(10,101,1);
  parabola(-101,0,2);
  parabola(-101,-10,3);
  parabola(-10,101,2);
  parabola(-101,10,3);
  parabola(-101,101,1);
  // c < 0
  parabola(0,10,-2);
  parabola(10,101,-2);
  parabola(-101,0,-3);
  parabola(-101,-10,-4);
  parabola(0,10,-5);
  parabola(10,101,-1);
  parabola(-101,0,-2);
  parabola(-101,-10,-3);
  parabola(-10,101,-2);
  parabola(-101,10,-3);
  parabola(-101,101,-5);
}

#define CTE 1000

// idem parabola
void parabola0(int lb, int ub, int c) { 
  int x=any_int(), x1=any_int(), x2=any_int(), r;

  // check test validity !
  verasco_assert(lb <= ub);

  // auxiliary init
  verasco_assume(-CTE <= x1 && x1 <= CTE);
  verasco_assume(-CTE <= x2 && x2 <= CTE);

  // initialization of inputs
  verasco_assume(lb <= x && x <= ub);
  verasco_assume(x==2*x1+3*x2);

  // TESTS
  r=c*x*(2*x1+3*x2);

  // Result check: best interval approx
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub); 
    } else if (ub <= 0) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else {
      verasco_assert(0 <= r);
      if (-lb <= ub) {
        verasco_assert(r <= c*ub*ub);
      } else {
        verasco_assert(r <= c*lb*lb);
      }
    }
  } else {
    if (0 <= lb) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else if (ub <= 0) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub);    
    } else {
      verasco_assert(r <= 0);
      if (-lb <= ub) {
        verasco_assert(c*ub*ub <= r);
      } else {
        verasco_assert(c*lb*lb <= r);      
      }
    }
  }

  // Result check: best poly approx on the "intern" side 
  if (c >= 0) {
    verasco_assert(r <= c*((ub+lb)*x - lb*ub));
  } else { 
    verasco_assert(c*((ub+lb)*x - lb*ub) <= r);
  }

  // Result check: on the "extern" side, tangent that 
  // generalizes the interval bound
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*(2*x-lb) <= r);
    } else if (ub <= 0) {
      verasco_assert(c*ub*(2*x-ub) <= r);
    } // else 0 <= r
  } else {
    if (0 <= lb) {
      verasco_assert(r <= c*lb*(2*x-lb));
    } else if (ub <= 0) {
      verasco_assert(r <= c*ub*(2*x-ub));
    } // else 0 <= r
  }

}

void test_parabola0() { 
  // 0 <= c 
  parabola0(0,10,2);
  // parabola0(10,101,2);  // ???
  parabola0(-101,0,3);
  // parabola0(-101,-10,4); // ???
  parabola0(0,10,5);
  parabola0(-101,0,2); 
  parabola0(-10,101,2); 
  parabola0(-101,10,3);
  parabola0(-101,101,1);
  // c < 0: TODO
  parabola0(0,10,-2);
  // parabola0(10,101,-2);  // ???
  parabola0(-101,0,-3);
  // parabola0(-101,-10,-4); // ???
  parabola0(0,10,-5);
  parabola0(-101,0,-2); 
  parabola0(-10,101,-2); 
  parabola0(-101,10,-3);
  parabola0(-101,101,-1);
}

// almost a parabola...
void parabola1(int lb, int ub, int c) { 
  int x=any_int(), y=any_int(), r;

  // check test validity !
  verasco_assert(lb <= ub);

  // initialization of inputs
  verasco_assume(lb <= x);
  verasco_assume(y <= ub);
  verasco_assume(x <= y);

  // TESTS
  r=c*x*y;

  // Result check: best interval approx
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub); 
    } else if (ub <= 0) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else {
      verasco_assert(c*lb*ub <= r);
      if (-lb <= ub) {
        verasco_assert(r <= c*ub*ub);
      } else {
        verasco_assert(r <= c*lb*lb);
      }
    }
  } else {
    if (0 <= lb) {
      verasco_assert(c*ub*ub <= r);
      verasco_assert(r <= c*lb*lb);
    } else if (ub <= 0) {
      verasco_assert(c*lb*lb <= r);
      verasco_assert(r <= c*ub*ub);    
    } else {
      if (-lb <= ub) {
        verasco_assert(c*ub*ub <= r);
      } else {
        verasco_assert(c*lb*lb <= r);      
      }
    }
  }

  // Result check: "intern" side 
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(r <= c*((ub+lb)*y - lb*ub));
    } else if (ub <= 0) {
      verasco_assert(r <= c*((ub+lb)*x - lb*ub));
    }
  } else {
    if (0 <= lb) {
      verasco_assert(c*((ub+lb)*y - lb*ub) <= r);
    } else if (ub <= 0) {
      verasco_assert(c*((ub+lb)*x - lb*ub) <= r);
    }
  }

  // Result check: "extern" side
  if (c >= 0) {
    if (0 <= lb) {
      verasco_assert(c*lb*(2*x-lb) <= r);
    } else if (ub <= 0) {
      verasco_assert(c*ub*(2*y-ub) <= r);
    }
  } else {
    if (0 <= lb) {
      verasco_assert(r <= c*lb*(2*x-lb));
    } else if (ub <= 0) {
      verasco_assert(r <= c*ub*(2*y-ub));
    }
  }
}

void test_parabola1() { 
  // 0 <= c 
  parabola1(0,10,2);
  parabola1(10,101,2);
  parabola1(-101,0,3);
  parabola1(-101,-10,4);
  parabola1(0,10,5);
  parabola1(10,101,1);
  parabola1(-101,0,2);
  parabola1(-10,101,2);
  parabola1(-101,10,3);
  parabola1(-101,101,1);
  // c < 0
  parabola1(0,10,-2);
  parabola1(10,101,-2);
  parabola1(-101,0,-3);
  parabola1(-101,-10,-4);
  parabola1(0,10,-5);
  parabola1(10,101,-1);
  parabola1(-101,0,-2);
  parabola1(-10,101,-2); 
  parabola1(-101,10,-3);
  parabola1(-101,101,-5);
}

int main(){
  test_parabola();
  test_parabola0();
  test_parabola1();
  return 0;
}
