(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

Require Import
  Coqlib Utf8 String PrintPos ToString
  Ctypes Integers
  Maps TreeAl Util AdomLib.

Module P := Tree_Properties( ZShareTree ).

Definition int_set : Type := ZShareTree.t unit.
Definition proj (i: int) : Z := Int.unsigned i.

Notation "x ∈ S" := (ZShareTree.get x S = Some tt): int_set_scope.

Local Open Scope int_set_scope.

Definition singleton (i: int) : int_set := ZShareTree.set (proj i) tt (ZShareTree.empty _).

Instance toString : ToString int_set :=
  { to_string x :=
  ("{ " ++
        ZShareTree.fold
        (fun s i _ => string_of_z i ++ "; " ++ s)
        x
        "}"
  )%string
  }.

Instance fs_gamma : gamma_op int_set int :=
  fun s i => proj i ∈ s.

Definition map (f: Z → Z) (s: int_set) : int_set :=
  ZShareTree.fold (fun m k _ => ZShareTree.set (f k) tt m) s (ZShareTree.empty unit).

Lemma map_in f (s: int_set) : ∀ x, x ∈ s → f x ∈ map f s.
Proof.
  unfold map.
  apply P.fold_rec.
  intros m m' a H H0 x.
  rewrite <- H. auto.
  intros x. repeat rewrite ZShareTree.gempty. intuition.
  intros m a k () H H0 H1 x.
  repeat rewrite ZShareTree.gsspec.
  repeat destruct ZShareTree.elt_eq. auto.
  apply False_ind. subst. intuition.
  auto.
  auto.
Qed.
