(* *********************************************************************)
(*                                                                     *)
(*              The Verasco static analyzer                            *)
(*                                                                     *)
(*          Vincent Laporte, Université de Rennes 1                    *)
(*          Jacques-Henri Jourdan, INRIA Paris-Rocquencourt            *)
(*                                                                     *)
(*  Copyright Institut National de Recherche en Informatique et en     *)
(*  Automatique and Université de Rennes 1.  All rights reserved.      *)
(*  This file is distributed under the terms of the GNU General        *)
(*  Public License as published by the Free Software Foundatio,        *)
(*  either version 2 of the License, or (at your option) any later     *)
(*  version.  This file is also distributed under the terms of the     *)
(*  INRIA Non-Commercial License Agreement.                            *)
(*                                                                     *)
(* *********************************************************************)

(* Heuristics for interval widening *)

open BinPos
open BinNums
open BinInt
open Datatypes
open Camlcoq
open Maps
open Cminor
open ToString

let hints =
  Zarith.([neg (shift_left Zarith.one 63);
          neg (shift_left Zarith.one 31);
          neg (shift_left Zarith.one 15);
          neg (shift_left Zarith.one 7);
          Zarith.zero; Zarith.one;
          sub (shift_left Zarith.one 7) Zarith.one;
          sub (shift_left Zarith.one 8) Zarith.one;
          sub (shift_left Zarith.one 15) Zarith.one;
          sub (shift_left Zarith.one 16) Zarith.one;
          sub (shift_left Zarith.one 31) Zarith.one;
          sub (shift_left Zarith.one 32) Zarith.one;
          sub (shift_left Zarith.one 63) Zarith.one])

let hints_rev = List.rev hints

exception Found of coq_Z

let next_larger (z : coq_Z) : coq_Z option =
  let z64 = Z.to_zarith z in
  try Some (Z.of_zarith (List.find (Zarith.lt z64) hints))
  with Not_found -> None

let next_smaller (z : coq_Z) : coq_Z option =
  let z64 = Z.to_zarith z in
  try Some (Z.of_zarith (List.find (Zarith.gt z64) hints_rev))
  with Not_found -> None

(* TODO *)
let next_larger_float f =
  Fappli_IEEE.B754_infinity false

(* TODO *)
let next_smaller_float f =
  Fappli_IEEE.B754_infinity true
